import json
import os
import uuid

import msgpack
from attrdict import AttrDict

from kodexa.mixins import registry
from kodexa.mixins.registry import get_mixin


class DocumentMetadata(dict):
    """
    A flexible dict based approach to capturing metadata for the document

    Sub-classes dict, and further allows attribute-like access to dictionary items.

        >>> d = DocumentMetadata({'a': 1})
        >>> d.a, d['a'], d.get('a')
        (1, 1, 1)
        >>> d.b = 2
        >>> d.b, d['b']
        (2, 2)
    """

    def __init__(self, *args, **kwargs):
        super(DocumentMetadata, self).__init__(*args, **kwargs)
        self.__dict__ = self


class ContentNode(object):
    """
    A Content Node identifies a section of the document containing logical
    grouping of information

    The node will have content and can include n number of features.

    You should always create a node using the Document create_node method to
    ensure that the correct mixins are applied.

        >>> new_page = document.create_node(type='page')
        <mango.model.model.ContentNode object at 0x7f80605e53c8>
        >>> current_content_node.add_child(new_page)

    or

        >>> new_page = document.create_node(type='page', content='This is page 1')
        <mango.model.model.ContentNode object at 0x7f80605e53c8>
        >>> current_content_node.add_child(new_page)
    """

    def __init__(self, document, type, content="", content_parts=[]):
        self.type = type
        self.content = content
        self.document = document
        self.content_parts = content_parts
        self.parent = None
        self.children = []
        self.index = 0
        self.uuid = str(uuid.uuid4())

        # Added for performance
        self._feature_map = {}

    def __str__(self):
        return f"ContentNode [type:{self.type}] ({len(self.get_features())} features, {len(self.children)} children) [" + str(
            self.content) + "]"

    def to_text(self):
        """
        Convert this node  structure into a text representation,
        which can be useful when trying to review the structure.


            >>> node.to_text()
        """
        return DefaultDocumentRender(self.document).node_to_text(self, 0)

    def _repr_html_(self):
        return self.to_html()

    def to_html(self):
        return DocumentRender(self.document).render_node(self)

    def to_mimetype(self):
        return DocumentRender(self.document).render_node_mimetype(self)

    def to_json(self):
        """
        Convert this node structure into a JSON object


            >>> node.to_json()
        """
        return json.dumps(self.to_dict())

    def to_dict(self):
        """
        Convert the ContentNode, and all its children into a simple dictionary

            >>> new_page = document.create_node(type='page')
            <mango.model.model.ContentNode object at 0x7f80605e53c8>
            >>> current_content_node.to_dict()
        """
        new_dict = {'type': self.type, 'content': self.content, 'content_parts': self.content_parts, 'features': [],
                    'index': self.index, 'children': []}
        for feature in self.get_features():
            new_dict['features'].append(feature.to_dict())

        for child in self.children:
            new_dict['children'].append(child.to_dict())
        return new_dict

    @staticmethod
    def from_dict(document, content_node_dict):
        new_content_node = document.create_node(type=content_node_dict['type'], content=content_node_dict[
            'content'] if 'content' in content_node_dict else None)

        if 'content_parts' in content_node_dict:
            new_content_node.content_parts = content_node_dict['content_parts']

        for dict_feature in content_node_dict['features']:
            new_feature = new_content_node.add_feature(dict_feature['name'].split(':')[0],
                                                       dict_feature['name'].split(':')[1],
                                                       dict_feature['value'], dict_feature['single'], True)
        for dict_child in content_node_dict['children']:
            new_content_node.add_child(ContentNode.from_dict(document, dict_child), dict_child['index'])
        return new_content_node

    def add_child(self, child, index=None):
        """
        Add a ContentNode as a child of this ContentNode

            >>> new_page = document.create_node(type='page')
            <mango.model.model.ContentNode object at 0x7f80605e53c8>
            >>> current_content_node.add_child(new_page)
        """
        if not index:
            child.index = len(self.children)
        else:
            child.index = index
        self.children.append(child)
        child.parent = self

    def get_children(self):
        """
        Returns a list of the children of this node

           >>> new_page = document.create_node(type='page')
           <mango.model.model.ContentNode object at 0x7f80605e53c8>
           >>> new_page.get_children()
           []
        """
        return self.children

    def set_feature(self, feature_type, name, value):
        """
        Sets a feature to this ContentNode, replacing the value

        You will need to provide the feature type, the name of the feature
        and then the value.

        Note this will replace any matching feature (i.e. with the same type and name)

           >>> new_page = document.create_node(type='page')
           <mango.model.model.ContentNode object at 0x7f80605e53c8>
           >>> new_page.add_feature('pagination','pageNum',1)
        """
        self.remove_feature(feature_type, name)
        return self.add_feature(feature_type, name, value)

    def add_feature(self, feature_type, name, value, single=True, serialized=False):
        """
        Add a new feature to this ContentNode.

        You will need to provide the feature type, the name of the feature
        and then the value.

        Note this will add a value to an existing feature, therefore
        the feature value might switch to being a list

           >>> new_page = document.create_node(type='page')
           <mango.model.model.ContentNode object at 0x7f80605e53c8>
           >>> new_page.add_feature('pagination','pageNum',1)
        """
        if self.has_feature(feature_type, name):
            feature = self.get_feature(feature_type, name)
            if feature.single:
                feature.single = False
            feature.value.append(value)
            return feature
        else:
            # Make sure that we treat the value as list all the time
            new_feature = ContentFeature(feature_type, name,
                                         [value] if single and not serialized else value, single)
            self._feature_map[new_feature.feature_type + ":" + new_feature.name] = new_feature
            return new_feature

    def get_feature(self, feature_type, name):
        """
        Gets the value for the given feature.

        You will need to provide the type and name of the feature. If no
        feature is find you will get None

           >>> new_page.get_feature('pagination','pageNum')
           1

        """
        return self._feature_map[feature_type + ":" + name] if feature_type + ":" + name in self._feature_map else None

    def get_features_of_type(self, feature_type):
        """
        Return a list of all the features of a specific type

           >>> new_page.get_features_of_type('tag')
           []

        :returns: list of the tags

        """
        return [i for i in self.get_features() if i.feature_type == feature_type]

    def has_feature(self, feature_type, name):
        """
        Determines if the feature with the given name and type exists on this content node

        You will need to provide the type and name of the feature. If the feature is present
        it will return True, else it will return False

           >>> new_page.has_feature('pagination','pageNum')
           True

        :returns: True if the feature is present

        """
        return feature_type + ":" + name in self._feature_map

    def get_features(self):
        """
        Returns a list of the features on this content node

        :returns: a list of the features present
        """
        return list(self._feature_map.values())

    def remove_feature(self, feature_type, name):
        """
        Determines if the feature with the given name and type exists on this content node

        You will need to provide the type and name of the feature. If the feature is present
        it will return True, else it will return False

           >>> new_page.remove_feature('pagination','pageNum')


        """
        results = self.get_feature(feature_type, name)
        if results:
            del self._feature_map[feature_type + ":" + name]

    def get_feature_value(self, feature_type, name):
        """
        Returns the assigned value for a given feature

        You will need to provide the type and name of the feature. If the feature is present
        it will return the value otherwise it will return None.

           >>> new_page.get_feature_value('pagination','pageNum')
           1

        """
        feature = self.get_feature(feature_type, name)

        # Need to make sure we handle the idea of a single value for a feature
        return None if feature is None else feature.value[0] if feature.single else feature.value

    def get_content(self):
        """
        Returns the content of the node

           >>> new_page.get_content()
           "This is page one"

        """
        return self.content

    def get_type(self):
        """
        Returns the type of the node

           >>> new_page.get_content()
           "page"

        """
        return self.type


class ContentFeature(object):
    """
    A feature that has been added to a ContentNode
    """

    def __init__(self, feature_type, name, value, description=None, single=True):
        self.feature_type = feature_type
        self.name = name
        self.value = value
        self.description = description
        self.single = single

    def __str__(self):
        return f"Feature [type='{self.feature_type}' name='{self.name}' value='{self.value}' single='{self.single}']"

    def to_dict(self):
        return {'name': self.feature_type + ':' + self.name, 'value': self.value, 'single': self.single}


class DocumentRender:

    def __init__(self, document):
        self.document = document
        self.default_renderer = DefaultDocumentRender(document)

    def to_text(self):
        return self.default_renderer.node_to_text(self.document.content_node, 0)

    def to_html(self):
        print(self.get_mixin_renderer())
        return self.get_mixin_renderer().to_html()

    def get_mixin_renderer(self):
        renderer_dict = registry.get_renderers(self.document)
        if renderer_dict:
            return next(iter(renderer_dict.values()))
        else:
            return self.default_renderer

    def render_node(self, node):
        return self.get_mixin_renderer().render_node(node)

    def to_mimetype(self):
        return self.get_mixin_renderer().to_mimetype()

    def render_node_mimetype(self, node):
        return self.get_mixin_renderer().render_node_mimetype(node)


class DefaultDocumentRender:

    def __init__(self, document):
        self.document = document

    def to_text(self):
        return self.node_to_text(self.document.content_node, 0)

    def to_html(self):
        return f"<h3>Document {self.document}</h3>"

    def render_node(self, node):
        out_list = []
        self.get_node_text_format(node, 0, out_list)
        return f"<h3>Content Node</h3><p>{self.node_to_text(node, 0)}"

    def node_to_text(self, node, level):
        out_list = []
        self.get_node_text_format(node, level, out_list)
        return os.linesep.join(out_list)

    def get_node_text_format(self, node, level, out_list):
        out_list.append("{}[type:{}]{}".format(" " * (4 * level), node.type, self.mixins_to_text(node)))

        if node.content_parts:
            print("Content Parts:")
            out_list.append("{}{}".format(" " * (4 * level), self.get_content_parts(node)))

        print("Content:")
        out_list.append("{}{}".format(" " * (4 * level), node.content))

        for child in node.get_children():
            self.get_node_text_format(child, level + 1, out_list)

    def get_content_parts(self, node):
        result = ""
        for part in node.content_parts:
            if isinstance(part, str):
                result = result + part
            # else:
            #     result = result + "#c" + str(part)

        return result

    def mixins_to_text(self, node):
        display = ""
        for mixin in self.document.get_mixins():
            if get_mixin(mixin):
                text = get_mixin(mixin).to_text(node)
                if text:
                    display = display + get_mixin(mixin).to_text(node)

        return display

    def to_mimetype(self):
        out_list = []
        self.get_node_text_format(self.document.content_node, 0, out_list)
        return self.prepare_mimetype_data('document', tuple(out_list))

    def render_node_mimetype(self, node):
        out_list = []
        self.get_node_text_format(node, 0, out_list)
        return self.prepare_mimetype_data('content_node', tuple(out_list))

    def prepare_mimetype_data(self, data_type, data_list):
        render_data = {'data_type': data_type, 'data': tuple(data_list)}
        bundle = {}
        bundle['application/vnd.mango.document+json'] = render_data
        return bundle


class Document(object):
    """
    A Document is a collection of metadata and a set of content nodes.
    """

    def __str__(self):
        return f"MDocument {self.uuid} {self.metadata}"

    def __init__(self, metadata: DocumentMetadata, content_node: ContentNode = None):
        self.metadata: DocumentMetadata = metadata
        self.content_node: ContentNode = content_node
        self.virtual: bool = False
        self._mixins = []
        self.uuid: str = str(uuid.uuid5(uuid.NAMESPACE_DNS, 'kodexa.com'))
        self.exceptions = []
        self.log = []
        self.version = "1.0.0"

        # Make sure we apply all the mixins
        registry.apply_to_document(self)

    def get_root(self):
        """
        Get the root content node for the document (same as content_node)


            >>> node = document.get_node()
        """
        return self.content_node

    def to_text(self):
        """
        Convert this document object structure into a text representation,
        which can be useful when trying to review the structure.


            >>> document.to_text()
        """
        return DocumentRender(self).to_text()

    def to_mdoc(self, file_path):
        """
        Write the document to the mdoc format (msgpack) which can be
        used with the Kodexa platform

            >>> document.to_mdoc('my-document.mdoc')

        :param file_path: the path to the mdoc you wish to create
        """
        with open(file_path, 'wb') as outfile:
            msgpack.pack(self.to_dict(), outfile, use_bin_type=True)

    @staticmethod
    def from_mdoc(file_path):
        """
        Read an mdoc file from the given file_path and

            >>> document = Document.from_mdoc('my-document.mdoc')

        :param file_path: the path to the mdoc file
        """
        with open(file_path, 'rb') as data_file:
            data_loaded = msgpack.unpack(data_file, raw=False)
        return Document.from_dict(data_loaded)

    def to_msgpack(self):
        """
        Convert this document object structure into a message pack


            >>> document.to_msgpack()
        """
        return msgpack.packb(self.to_dict(), use_bin_type=True)

    def to_json(self):
        """
        Convert this document object structure into a JSON object


            >>> document.to_json()
        """
        return json.dumps(self.to_dict(), ensure_ascii=False)

    def _repr_html_(self):
        return self.to_html()

    def to_html(self):
        return DocumentRender(self).to_html()

    def to_dict(self):
        """
        Convert this document object structure into a simple set of dictionaries


            >>> document.to_dict()
        """
        return {'version': self.version, 'metadata': self.metadata,
                'content_node': self.content_node.to_dict() if self.content_node else None,
                'mixins': self._mixins,
                'exceptions': self.exceptions,
                'log': self.log,
                'uuid': self.uuid}

    def to_mimetype(self):
        return DocumentRender(self).to_mimetype()

    @staticmethod
    def from_dict(doc_dict):
        """
        Build a new document from a dictionary


            >>> Document.from_dict(doc_dict)
        """
        new_document = Document(DocumentMetadata(doc_dict['metadata']))
        for mixin in doc_dict['mixins']:
            registry.add_mixin_to_document(mixin, new_document)
        new_document.version = doc_dict['version'] if 'version' in doc_dict else '1.0.0'
        new_document.log = doc_dict['log'] if 'log' in doc_dict else []
        new_document.exceptions = doc_dict['exceptions'] if 'exceptions' in doc_dict else []
        new_document.uuid = doc_dict['uuid'] if 'uuid' in doc_dict else str(
            uuid.uuid5(uuid.NAMESPACE_DNS, 'kodexa.com'))
        if 'content_node' in doc_dict and doc_dict['content_node']:
            new_document.content_node = ContentNode.from_dict(new_document, doc_dict['content_node'])

        return new_document

    @staticmethod
    def from_json(json_string):
        """
        From a JSON string create an instance of a Document

            >>> document.from_json()
        """
        return Document.from_dict(json.loads(json_string, object_hook=AttrDict))

    @staticmethod
    def from_msgpack(bytes):
        """
        From a message pack byte array create an instance of a Document

            >>> document.from_msgpack()
        """
        return Document.from_dict(msgpack.unpackb(bytes, raw=False))

    def get_mixins(self):
        """
        Get the list of mixins that have been enabled on this document.


            >>> document.get_mixins()
            ['spatial','finders']
        """
        return self._mixins

    def add_mixin(self, mixin):
        """
        Add the given mixin to this document,  this will apply the mixin to all the content nodes,
        and also register it with the document so that future invocations of create_node will ensure
        the node has the mixin appled.

            >>> document.add_mixin('spatial')
        """
        registry.add_mixin_to_document(mixin, self)

    def create_node(self, type=type, content=None, virtual=False, parent=None, index=0):
        """
        Creates a new node for the document, this doesn't add the node to the document however
        it does ensure that any mixins that have been applied to the document will also be
        available on the new node

            >>> document.create_node(type='page')
            <mango.model.model.ContentNode object at 0x7f80605e53c8>
        """
        content_node = ContentNode(document=self, type=type, content=content)
        content_node.parent = parent
        content_node.index = index
        content_node.virtual = virtual
        registry.add_mixins_to_document_node(self, content_node)
        if virtual:
            for mixin_name in self.get_mixins():
                mixin = registry.get_mixin(mixin_name)
                add_features_to_virtual_node = getattr(mixin, "add_features_to_virtual_node", None)
                if callable(add_features_to_virtual_node):
                    add_features_to_virtual_node(content_node)
        return content_node



import itertools
import re
from enum import Enum

from attrdict import AttrDict

from kodexa.mixins.util import add_method_to_node


class FindDirection(Enum):
    CHILDREN = 1
    PARENT = 2


class Traverse(Enum):
    SIBLING = 1
    CHILDREN = 2
    PARENT = 3
    ALL = 4


def get_all_content(self, separator=" "):
    """
    This will build the complete content, including the content of children.

        >>> document.content_node.get_all_content()
        "This string is made up of multiple nodes"

    :param separator: the separate to use to join the content together (default is " ")
    """
    s = ""
    if self.get_content():
        s += self.get_content()
        s += separator
    for child in self.children:
        s += child.get_all_content(separator)
        s += separator
    return s.strip()


def move_child_to_parent(self, target_child, target_parent):
    """
    This will move the target_child, which must be a child of the node, to a new parent.

    It will be added to the end of the parent

        >>> document.content_root.move_child_to_parent(document.content_root.find(type_ref='line'), document.content_root)

    :param target_child: the child node that needs to be moved
    :param target_parent: the parent to attach this node to
    """
    self.children.remove(target_child)
    target_parent.add_child(target_child)


def adopt_children(self, children, replace=False):
    """
    This will take a list of content nodes and adopt them under this node, ensuring they are
    re-parented.

    It will be added to the end of the parent

        >>> document.content_root.adopt_children(document.content_root.find(type_ref='line'), replace=True)

    :param children: a list of the children to adopt
    :param replace: if True the node will remove all current children and replace them with the new list
    """

    if replace:
        for child in self.children:
            self.parent = None
        self.children = []

    for child in children:
        self.add_child(child)


def remove_tag(self, tag_name):
    """
        This will remove a tag from the given node

             >>> document.content_node.remove_tag(tag_name='foo')

        :param tag_name: The tag to be applied to the nodes in range
        """
    self.remove_feature('tag', tag_name)


def collect_nodes_to(self, end_node):
    """
    Return a list of the sibling nodes between the current node and the end node

         >>> document.content_node.children[0].collect_nodes_to(end_node=document.content_node.children[5])

    :param end_node: The node to end at
    """
    nodes = []
    current_node = self
    while current_node.uuid != end_node.uuid:
        nodes.append(current_node)
        if current_node.has_next_node():
            current_node = current_node.next_node()
        else:
            break
    return nodes


def tag_nodes_to(self, end_node, tag_name):
    """
    Tag all the nodes from this node to the end node with the given tag name

          >>> document.content_node.children[0].tag_nodes_to(document.content_node.children[5], tag_name='foo')

     :param end_node: The node to end with
     :param tag_name: The tag name
     """
    [node.tag(tag_name) for node in self.collect_nodes_to(end_node)]


def tag_range(self, start_content_re, end_content_re, tag_name, type_re='.*', use_all_content=False):
    """
    This will tag all the child nodes between the start and end content regular expressions

         >>> document.content_node.tag_range(start_content_re='.*Cheese.*', end_content_re='.*Fish.*', tag_name='foo')

    :param start_content_re: The regular expression to match the starting child
    :param end_content_re: The regular expression to match the ending child
    :param tag_name: The tag to be applied to the nodes in range
    :param type_re: The type to match (default is all)
    :param use_all_content: Use full content (including child nodes, default is False)
    """

    # Could be line, word, or content-area
    all_nodes = self.findall(type_re=type_re)

    start_index_list = [n_idx for n_idx, node in enumerate(all_nodes)
                        if re.compile(start_content_re).match(node.get_all_content()
                                                              if use_all_content else node.content)]
    end_index_list = [n_idx for n_idx, node in enumerate(all_nodes)
                      if re.compile(end_content_re).match(node.get_all_content()
                                                          if use_all_content else node.content)]

    start_index = 0 if start_content_re == '' else \
        start_index_list[0] if len(start_index_list) > 0 else None
    if start_index is not None:
        end_index_list = [i for i in end_index_list if i >= start_index]

    end_index = len(all_nodes) if end_content_re == '' else \
        end_index_list[0] if len(end_index_list) > 0 else len(all_nodes)

    if start_index is not None:
        [node.tag(tag_name) for node in all_nodes[start_index:end_index]]


def tag(self, tag_name, type_re=None, content_re=None,
        use_all_content=False, node_only=False, include_children=False,
        fixed_position=None):
    """
    This will tag (see Feature Tagging) the expression groups identified by the regular expression.

        >>> document.content_node.find(content_re='.*Cheese.*').tag('is_cheese')

    :param tag_name: the name of the tag to apply
    :param type_re: regular expression to make the type (default .*)
    :param content_re: the regular expression that you wish to use to tag, note that we will create a tag for each matching group
    :param use_all_content: apply the regular expression to the all_content (include content from child nodes)
    :param node_only: Ignore the matching groups and tag the whole node
    :param include_children: Include recurse into children and tag where matching
    :param fixed_position: use a fixed position, supplied as a tuple i.e. - (4,10) tag from position 4 to 10 (default None)
    """

    type_match = True

    if fixed_position:
        self.add_feature('tag', tag_name,
                         Tag(fixed_position[0], fixed_position[1], self.content[fixed_position[0]:fixed_position[1]]))
    else:
        if type_re:
            type_re_compiled = re.compile(type_re)
            type_match = type_re_compiled.match(self.type)

        if type_match:
            if not content_re:
                self.add_feature('tag', tag_name, [])
            else:
                pattern = re.compile(content_re)
                if not use_all_content:
                    if self.content:
                        content = self.content
                    else:
                        return
                else:
                    content = self.get_all_content()

                match = pattern.match(content)
                if match:
                    if node_only:
                        self.add_feature('tag', tag_name, Tag())
                    else:
                        for index, m in enumerate(match.groups()):
                            idx = index + 1
                            self.add_feature('tag', tag_name, Tag(match.start(idx), match.end(idx), match.group(idx)))

    if include_children:
        for child in self.children:
            child.tag(tag_name, type_re, content_re, use_all_content, node_only, include_children, fixed_position)


def get_tags(self):
    """
    Returns a list of the names of the tags on the given node

        >>> document.content_node.find(content_re='.*Cheese.*').get_tags()
        ['is_cheese']

    :return: A list of the tag name
    """
    return [i.name for i in self.get_features_of_type("tag")]


def get_tag(self, tag_name):
    """
    Returns the value of a tag, this can be either a single list [start,end,value] or if multiple parts of the
    content of this node match you can end up with a list of lists i.e. [[start1,end1,value1],[start2,end2,value2]]

        >>> document.content_node.find(content_re='.*Cheese.*').get_tag('is_cheese')
        [0,10,'The Cheese Moved']

    :param tag_name: The name of the tag

    :return: The tagged location and value (or a list if more than one)
    """
    return self.get_feature_value('tag', tag_name)


def get_all_tags(self):
    """
    Returns a list of the names of the tags on the given node and all its children

        >>> document.content_node.find(content_re='.*Cheese.*').get_all_tags()
        ['is_cheese']

    :return: A list of the tag names
    """
    tags = []
    tags.extend(self.get_tags())
    for child in self.children:
        tags.extend(child.get_all_tags())
    return list(set(tags))


def has_tags(self):
    """
    Returns True if the node has any tags

        >>> document.content_node.find(content_re='.*Cheese.*').has_tags()
        True

    :return: True if node has any tags else False
    """
    return len([i.value for i in self.get_features_of_type("tag")]) > 0


def has_tag(self, tag):
    """
    Returns True if the node has given tag

        >>> document.content_node.find(content_re='.*Cheese.*').has_tag('is_cheese')
        True
        >>> document.content_node.find(content_re='.*Cheese.*').has_tag('is_fish')
        False

    :param tag: The tag name

    :return: True if node has tag else False
    """
    for feature in self.get_features():
        if feature.feature_type == 'tag' and feature.name == tag:
            return True
    return False


def update_content_markup(self):
    """
    This will tag (see Feature Tagging) the expression groups identified by the regular expression
    """


def find(self, content_re=".*", type_re=".*", direction=FindDirection.CHILDREN, tag_name=None, instance=0,
         tag_name_re=None, use_all_content=False):
    """
    Search for a node that matches on the value and or type using regular expressions

        >>> document.content_node.find(content_re='.*Cheese.*',instance=2)
        <mango.model.model.ContentNode object at 0x7f80605e53c8>

    :param content_re: the regular expression to match the nodes content (default '.*')
    :param type_re: the regular expression to match the nodes type (default '.*')
    :param direction: the direction to search, either FindDirection.CHILDREN or FindDirection.PARENT (default CHILDREN)
    :param tag_name: the tag name that must exist
    :param instance: the instance to return (0=first)
    :param tag_name_re: the regular expression to match for the tag_name
    :param use_all_content: match the content for all child nodes to (default False)

    :return: either an instance of ContentNode (if found), or None
    """
    results = self.findall(content_re, type_re, direction, tag_name, tag_name_re, use_all_content)
    if len(results) < instance + 1:
        return None
    else:
        return results[instance]


def find_with_feature_value(self, feature_type, feature, value, direction=FindDirection.CHILDREN, instance=1):
    """
    Search for a node with a specific feature type, name and value

        >>> document.content_node.find_with_feature_value(feature_type='tag',feature='is_cheese',value=[1,10,'The Cheese has moved'])
        <mango.model.model.ContentNode object at 0x7f80605e53c8>

    :param feature_type: the feature type
    :param feature: the feature name
    :param value: the feature value
    :param direction: the direction to search, either FindDirection.CHILDREN or FindDirection.PARENT (default CHILDREN)
    :param instance: the instance to get (defaul 1)

    :return: either an instance of ContentNode (if found), or None
    """
    return next(
        itertools.islice(self.findall_with_feature_value(feature_type, feature, value, direction), instance - 1, 1),
        None)


def findall_with_feature_value(self, feature_type, feature_name, value, direction=FindDirection.CHILDREN):
    """
    Search for all nodes with a specific feature type, name and value

        >>> document.content_node.findall_with_feature_value(feature_type='tag',feature='is_cheese', value=[1,10,'The Cheese has moved'])
        [<mango.model.model.ContentNode object at 0x7f80605e53c8>]

    :param feature_type: the feature type
    :param feature_name: the feature name
    :param value: the feature value
    :param direction: the direction to search, either FindDirection.CHILDREN or FindDirection.PARENT (default CHILDREN)

    :return: list of the matching content nodes
    """
    if self.has_feature(feature_type, feature_name) and value == self.get_feature_value(feature_type, feature_name):
        yield self

    if direction is FindDirection.CHILDREN:
        for child in self.get_children():
            yield from child.findall_with_feature_value(feature_type, feature_name, value, direction)
    else:
        if self.parent:
            yield from self.parent.findall_with_feature_value(feature_type, feature_name, value, direction)


def is_first_child(self):
    """
    Returns True if this is the first child, also True if it has no parent

    :return: True if this is the first child
    """

    if not self.parent:
        return True
    else:
        return self.index == 0


def is_last_child(self):
    """
    Returns True if this is the last child, also True if it has no parent

    :return: True if this is the first child
    """

    if not self.parent:
        return True
    else:
        return self.index == self.parent.get_last_child_index()


def get_last_child_index(self):
    """
    Returns the max index value for the children of this node, if the node has no children
    it returns None

    :return: Thhe max index of the children of this node, or None is no children
    """

    if not self.children:
        return None

    max_index = 0
    for child in self.children:
        if child.index > max_index:
            max_index = child.index

    return max_index


def get_node_at_index(self, index):
    """
    Returns the node at a specific index,  if the index it outside the first (0), or
    last index it will return null, if not it will return the node at index, if there
    isn't a node at that index it will return a 'virtual' node that will represent
    the node to the side of this node and have an index and no features or content

    :param index: The index (zero-based) for the child node

    :return: Node at index, or None is the index is outside the boundaries of child nodes
    """
    if self.children:

        if index < self.children[0].index:
            virtual_node = self.document.create_node(type=self.children[0].type, virtual=True, parent=self,
                                                     index=index)
            return virtual_node

        last_child = None
        for child in self.children:
            if child.index < index:
                last_child = child
            elif child.index == index:
                return child
            else:
                break

        if last_child:
            if last_child.index is not index and index < self.children[-1].index:
                virtual_node = self.document.create_node(type=last_child.type, virtual=True, parent=self,
                                                         index=index)
                return virtual_node
        else:
            return None
    else:
        return None


def has_next_node(self, type_re=".*", skip_virtual=False):
    """
    Returns True if the node has a next node

    :param type_re: Type name (regular expression)
    :param skip_virtual: True to skip any virtual nodes and only return the next real node (default False)

    :return: True if there is a next sibling node
    """
    return self.next_node(type_re, skip_virtual=skip_virtual) is not None


def has_previous_node(self, type_re=".*", skip_virtual=False):
    """
     Returns True if the node has a previous node

    :param type_re: Type name (regular expression)
    :param skip_virtual: True to skip any virtual nodes and only return the next real node (default False)

     :return: True if there is a previous sibling node
     """
    return self.previous_node(type_re=type_re, skip_virtual=skip_virtual) is not None


def next_node(self, type_re='.*', skip_virtual=False, has_no_content=False, traverse=Traverse.SIBLING):
    """
    Returns the next sibling content node, note that this logic is based on the index,
    therefore the next node might actually be a virtual node that is created to fill a gap
    in the document,  since the index allows for sparse documents

    :param type_re: the regular expression for the type of node (default '.*')
    :param skip_virtual: True to skip any virtual nodes and only return the next real node (default False)
    :param has_no_content: True if you only want to return a node that has no content
    :param traverse: By default we traverse siblings, however you can include CHILDREN, PARENT or ALL

    :return: the next node or None if no node exists
    """
    search_index = self.index + 1
    compiled_type_re = re.compile(type_re)

    while True:
        node = self.parent.get_node_at_index(search_index)

        if not node:
            return node

        if compiled_type_re.match(node.type) and (not skip_virtual or not node.virtual):
            if (not has_no_content) or (has_no_content and not node.content):
                return node
        if not node:
            return node
        search_index += 1


def previous_node(self, type_re='.*', skip_virtual=False, has_no_content=False, traverse=Traverse.SIBLING):
    """
    Returns the previous sibling content node, note that this logic is based on the index,
    therefore the next node might actually be a virtual node that is created to fill a gap
    in the document,  since the index allows for sparse documents

    :param type_re: the regular expression for the type of node (default '.*')
    :param skip_virtual: True to skip any virtual nodes and only return the next real node (default False)
    :param has_no_content: True if you only want to return a node that no content
    :param traverse: By default we traverse siblings, however you can include CHILDREN, PARENT or ALL

    :return: the previous node or None if no node exists
    """
    if self.index == 0:
        if traverse == traverse.ALL or traverse == traverse.PARENT and self.parent:
            # Lets look for a previous node on the parent
            return self.parent.previous_node(type_re, skip_virtual, has_no_content, traverse)
        else:
            return None

    search_index = self.index - 1
    compiled_type_re = re.compile(type_re)

    while True:
        node = self.parent.get_node_at_index(search_index)

        if not node:
            return node

        if compiled_type_re.match(node.type) and (not skip_virtual or not node.virtual):
            if (not has_no_content) or (has_no_content and not node.content):
                return node

        search_index -= 1


def findall(self, content_re=".*", type_re=".*", direction=FindDirection.CHILDREN, tag_name=None,
            tag_name_re=None, use_all_content=False):
    """
    Search for nodes that matches on the value and or type using regular expressions

        >>> document.content_node.findall(content_re='.*Cheese.*')
        [<mango.model.model.ContentNode object at 0x7f80605e53c8>,
        <mango.model.model.ContentNode object at 0x7f80605e53c8>]

    :param content_re: the regular expression to match the nodes content (default '.*')
    :param type_re: the regular expression to match the nodes type (default '.*')
    :param direction: the direction to search, either FindDirection.CHILDREN or FindDirection.PARENT (default CHILDREN)
    :param tag_name: the tag name that must exist
    :param tag_name_re: the regular expression to match for the tag_name
    :param use_all_content: match the content for all child nodes to (default False)

    :return: list of matching content nodes


    """
    value_compiled = re.compile(content_re)
    type_compiled = re.compile(type_re)
    if tag_name_re:
        tag_name_re_compiled = re.compile(tag_name_re)
    else:
        tag_name_re_compiled = None
    return self.findall_compiled(value_compiled, type_compiled, direction, tag_name, tag_name_re_compiled,
                                 use_all_content)


def findall_compiled(self, value_re_compiled, type_re_compiled, direction, tag_name, tag_name_compiled,
                     use_all_content):
    """
    Search for a node that matches on the value and or type using
    regular expressions using compiled expressions
    """
    hits = []

    if use_all_content:
        content = self.get_all_content()
    else:
        content = "" if not self.get_content() else self.get_content()

    if value_re_compiled.match(content) and type_re_compiled.match(self.get_type()):

        if tag_name_compiled:
            for tag_name in self.get_tags():
                if tag_name_compiled.match(tag_name):
                    hits.append(self)
                    break

        elif not tag_name or self.has_tag(tag_name):
            hits.append(self)

    if direction is FindDirection.CHILDREN:
        for child in self.get_children():
            hits.extend(child.findall_compiled(value_re_compiled, type_re_compiled, direction, tag_name,
                                               tag_name_compiled, use_all_content))
    else:
        if self.parent:
            hits.extend(self.parent.findall_compiled(value_re_compiled, type_re_compiled, direction, tag_name,
                                                     tag_name_compiled, use_all_content))

    return hits


class CoreMixin:

    @staticmethod
    def get_name():
        return "core"

    @staticmethod
    def get_renderer(document):
        return None

    @staticmethod
    def to_text(self):
        return "[tags:" + ",".join(self.get_tags()) + "]"

    @staticmethod
    def apply_to(node):
        add_method_to_node(get_all_content, node)
        add_method_to_node(get_all_tags, node)
        add_method_to_node(get_tag, node)
        add_method_to_node(find, node)
        add_method_to_node(tag, node)
        add_method_to_node(tag_range, node)
        add_method_to_node(has_tag, node)
        add_method_to_node(has_tags, node)
        add_method_to_node(findall, node)
        add_method_to_node(findall_compiled, node)
        add_method_to_node(find_with_feature_value, node)
        add_method_to_node(findall_with_feature_value, node)
        add_method_to_node(get_tags, node)
        add_method_to_node(move_child_to_parent, node)
        add_method_to_node(adopt_children, node)
        add_method_to_node(remove_tag, node)
        add_method_to_node(collect_nodes_to, node)
        add_method_to_node(tag_nodes_to, node)
        add_method_to_node(get_node_at_index, node)
        add_method_to_node(has_next_node, node)
        add_method_to_node(has_previous_node, node)
        add_method_to_node(next_node, node)
        add_method_to_node(previous_node, node)
        add_method_to_node(get_last_child_index, node)
        add_method_to_node(is_first_child, node)
        add_method_to_node(is_last_child, node)


class Tag(AttrDict):

    def __init__(self, start=None, end=None, value=None):
        self.start = start
        self.end = end
        self.value = value



import json
import os

import msgpack
import requests
from requests.auth import HTTPBasicAuth

from kodexa import get_source, Document


class KodexaService:
    """
    Allows you to interact with a content service that has been deployed in the Kodexa cloud
    """

    def __init__(self, organization, service, version=None, attach_source=False, options={}, auth=[],
                 cloud_url='https://cloud.kodexa.com'):
        self.organization = organization
        self.service = service
        self.version = version
        self.attach_source = attach_source
        self.options = options
        self.auth = auth
        self.cloud_url = cloud_url

    def get_name(self):
        return f"Kodexa Content Service ({self.organization}/{self.service})"

    def process(self, document):
        # We use a very simple approach to get the document to the service

        service_message = {
            "document": document.to_dict(),
            "options": self.options,
            "stores": {}
        }

        upload_files = {}

        data = {
            "options": json.dumps(self.options)
        }

        if self.attach_source:
            upload_files["file"] = get_source(document)
        else:
            upload_files["message"] = msgpack.packb(service_message, use_bin_type=True)

        env_user = os.getenv('KODEXA_USER')
        env_password = os.environ.get('KODEXA_PASSWORD')

        if len(self.auth) == 0 and env_user and env_password:
            self.auth = [env_user, env_password]

        endpoint = f"{self.cloud_url}/api/processing/{self.organization}/{self.service}/upload"

        r = requests.post(endpoint, files=upload_files, data=data, auth=HTTPBasicAuth(self.auth[0], self.auth[1]))
        response_message = json.loads(r.text)
        return Document.from_dict(response_message['document'])



import re

from kodexa.mixins.util import add_method_to_node



class NavigationMixin:

    @staticmethod
    def get_name():
        return "navigation"

    @staticmethod
    def get_renderer(document):
        return None

    @staticmethod
    def to_text(self):
        return None

    @staticmethod
    def get_dependencies():
        return ['core']

    @staticmethod
    def apply_to(node):
        pass
